/* 
 * original code by Rick Gesser
 * Contributor(s): 
 * bernd.mielke@snafu.de
 */
#include <iostream.h>
#include <fstream.h>
#include "nsString.h"
#include "prio.h"
#include "plstr.h"
#include "CScanner.h"
#include "CCPPTokenizer.h"
#include "tokens.h"
#include "CToken.h"

//*******************************************************************************
// Global variables...
//*******************************************************************************
bool  gVerbose=false;
bool  gShowIncludeErrors=false;
bool  gEmitHTML=false;
bool  gSloppy=false;
const char *gRootDir;


class IPattern {
public:
  virtual CToken* scan(CCPPTokenizer& aTokenizer,int anIndex,ostream& anOutput)=0;
};


class CPattern {
public:
  CPattern(const nsCString& aFilename) : mNames(0), mFilename(aFilename), mSafeNames(0) {
    mIndex=0;
    mLineNumber=1;
    mErrCount=0;
  }

  CToken* matchID(nsDeque& aDeque,nsCString& aName){
    int theMax=aDeque.GetSize();
    for(int theIndex=0;theIndex<theMax;theIndex++){
      CToken* theToken=(CToken*)aDeque.ObjectAt(theIndex);
      if(theToken){
        nsCString& theName=theToken->getStringValue();
        if(theName==aName) {
          return theToken;
        }
      }
    }
    return 0;
  }

  /***************************************************************************
    This gets called when we know we've seen an if statement (at anIndex).
    Look into the parm list (...) for an ID. If you find one, then you know
    it's a conditioned identifier which can be safely ignored during other
    error tests.
   ***************************************************************************/
  CToken* findSafeIDInIfStatement(CCPPTokenizer& aTokenizer) {
    CToken* result=0;

    enum eState {eSkipLparen,eFindID};
    eState  theState=eSkipLparen;
    CToken* theToken=0;
    while((theToken=aTokenizer.getTokenAt(++mIndex))) {
      int   theType=theToken->getTokenType();
      nsCString& theString=theToken->getStringValue();
      PRUnichar theChar=theString.GetBuffer()[0];

      switch(theState) {
        case eSkipLparen:
            //first scan for the lparen...
          switch(theType) {
            case eToken_comment:
            case eToken_whitespace:
            case eToken_newline:
              mLineNumber+=theString.CountChar('\n'); break;

            case eToken_operator:
              switch(theChar) {
                case '(': theState=eFindID; break;
                case ')': return 0; break;
                default:
                  break;
              } //switch
              break;
            default:
              break;
          } //switch
          break;

          //now scan for the identifier...
        case eFindID:
          switch(theType) {
            case eToken_comment:
            case eToken_whitespace:
            case eToken_newline:
              mLineNumber+=theString.CountChar('\n'); break;

            case eToken_identifier:
              if(matchID(mNames,theString)) {
                return theToken;
              }
              break;
            case eToken_operator:
              switch(theChar) {
                case ')':
                case ';':
                case '{':
                case '(': 
                  return result;
                  break;
                default:
                  break;
              } //switch
              break;
            default:
              break;
          } //switch
          break;

        default:
          switch(theType) {
            case eToken_comment:
            case eToken_whitespace:
            case eToken_newline:
              mLineNumber+=theString.CountChar('\n'); 
              break;
            default:
              break;
          }//switch
          break;
      } //switch
    }
    return result;
  }


  /***************************************************************************
    This gets called when we know we've seen an NS_ENSURE_TRUE statement (at anIndex).
    Look into the parm list (...) for an ID. If you find one, then you know
    it's a conditioned identifier which can be safely ignored during other
    error tests.
   ***************************************************************************/
  CToken* findSafeIDInEnsureTrue(CCPPTokenizer& aTokenizer) {

	CToken* result=0;
    enum eState {eSkipLparen,eFindID};
    eState  theState=eSkipLparen;
    CToken* theToken=0;
    while((theToken=aTokenizer.getTokenAt(++mIndex))) {
      int   theType=theToken->getTokenType();
      nsCString& theString=theToken->getStringValue();
      PRUnichar theChar=theString.GetBuffer()[0];

      switch(theState) {
        case eSkipLparen:
            //first scan for the lparen...
          switch(theType) {
            case eToken_comment:
            case eToken_whitespace:
            case eToken_newline:
              mLineNumber+=theString.CountChar('\n'); break;

            case eToken_operator:
              switch(theChar) {
                case '(': theState=eFindID; break;
                case ')': return 0; break;
                default:
                  break;
              } //switch
              break;
            default:
              break;
          } //switch
          break;

          //now scan for the identifier...
        case eFindID:
          switch(theType) {
            case eToken_comment:
            case eToken_whitespace:
            case eToken_newline:
              mLineNumber+=theString.CountChar('\n'); break;

            case eToken_identifier:
              if(matchID(mNames,theString)) {
                return theToken;
              }
              break;
            case eToken_operator:
              switch(theChar) {
                case ')':
                case ';':
                case '{':
                case '(': 
                  return result;
                  break;
                default:
                  break;
              } //switch
              break;
            default:
              break;
          } //switch
          break;

        default:
          switch(theType) {
            case eToken_comment:
            case eToken_whitespace:
            case eToken_newline:
              mLineNumber+=theString.CountChar('\n'); 
              break;
            default:
              break;
          }//switch
          break;
      } //switch
    }
    return result;
  }
   /***************************************************************************
    We're looking for the variable that is new created by a new 
   ***************************************************************************/
  CToken* findNewVariable(CCPPTokenizer& aTokenizer, int anIndex){
	  CToken* theToken=0;
	  while((theToken=aTokenizer.getTokenAt(--anIndex))){
		  if(theToken) {
			  int     theType=theToken->getTokenType();
			   nsCString& theString=theToken->getStringValue();
			  switch(theType){
				 case eToken_semicolon:
					return 0;
					break;
				 case eToken_operator:
					 if((0!=theString.CompareWithConversion("=",true,1))){
						 return 0;
					 }
					 break;
				 case eToken_identifier:
					 if((0==theString.CompareWithConversion("return",true,6))){
						 return 0;
					 }
					 else {
					 return theToken;
					 }
					 break;
				 default:
					 break;
			  }
		  } //if
	  } //while
	  return 0;
  }
  void check_unsafe_ID(CToken* theToken){
  }
  void check_save_ID(CToken* theToken){
  }
  /***************************************************************************
    We're looking for the pattern foo-> without the preceeding (conditional) if():
    Assume that the given id (aName) exists in the tokenizer at anIndex.

    if(foo)
      foo->
   ***************************************************************************/
  CToken* findPtrDecl(CCPPTokenizer& aTokenizer,int anIndex,PRUnichar& anOpChar,bool allocRequired) {

    enum eState {eFindNew,eFindType,eFindPunct,eFindStar};
    CToken* theToken=0;
    CToken* theID=0;
    eState  theState=eFindStar;
    
    while((theToken=aTokenizer.getTokenAt(++anIndex))){
      if(theToken) {
        nsCString& theString=theToken->getStringValue();
        int     theType=theToken->getTokenType();
        
        anOpChar=theString.GetBuffer()[0];
        switch(theType){

          case eToken_semicolon:
          case eToken_operator:
            switch(theState) {
              case eFindStar:
                if('*'==anOpChar)
                  theState=eFindType;
                else return 0;
                break;
              case eFindPunct:
                switch(anOpChar) {
                  case '=':
                    if(!allocRequired)
                      return theID;
                    else theState=eFindNew;
                    break;
                  case ';':
                    return (allocRequired) ? 0 : theID;
                    break;
                  default:
                    break;
                }
                break;
              default:
                return 0;
            }
            break;

          case eToken_identifier:
            switch(theState){
              case eFindType:
                theID=theToken;
                theState=eFindPunct;
                break;
              case eFindNew:
                if(theString.Equals("new") || theString.Equals("PR_Malloc"))
                  return theID;
                else return 0;
                break;
              default:
                return 0;
          }
          break;

          default:
            break;
        }
      }//if
    }//while
    return 0;
  }

  /***************************************************************************
    We're looking for the pattern *foo.
   ***************************************************************************/
  bool hasSimpleDeref(CCPPTokenizer& aTokenizer,int anIndex) {
    bool result=false;

    int theMin=anIndex-3;
    for(int theIndex=anIndex-1;theIndex>theMin;theIndex--){
      CToken* theToken=aTokenizer.getTokenAt(theIndex);
      if(theToken) {
        nsCString& theString=theToken->getStringValue();
        int     theType=theToken->getTokenType();
        PRUnichar   theChar=theString.GetBuffer()[0];
        if(eToken_operator==theType) {
          if('*'==theChar){
            result=true;
          }
        }
      }
    }

    if(result) {
      int theMax=anIndex+3;
      for(int theIndex=anIndex;theIndex<theMax;theIndex++){
        CToken* theToken=aTokenizer.getTokenAt(theIndex);
        if(theToken) {
          nsCString& theString=theToken->getStringValue();
          int     theType=theToken->getTokenType();
          PRUnichar   theChar=theString.GetBuffer()[0];
          if(eToken_operator==theType) {
            if((','==theChar) || (')'==theChar)){
              result=false;
            }
          }
        }
      }
    }

    return result;
  }

  /***************************************************************************
    We're looking for the pattern foo-> without the preceeding (conditional) if():
    Assume that the given id (aName) exists in the tokenizer at anIndex.

    if(foo)
      foo->
   ***************************************************************************/
  bool hasUnconditionedDeref(nsCString& aName,CCPPTokenizer& aTokenizer,int anIndex) {
    bool result=false;

    enum eState {eFailure,eFindIf,eFindDeref};
    eState  theState=eFindDeref;
    int     theIndex=anIndex+1;
    
    while(eFailure!=theState){
      CToken* theToken=aTokenizer.getTokenAt(theIndex);
      if(theToken) {
        nsCString& theString=theToken->getStringValue();
        int     theType=theToken->getTokenType();
        PRUnichar   theChar=theString.GetBuffer()[0];
        switch(theState){
          case eFindDeref:
              switch(theType) {
                case eToken_whitespace:
                case eToken_comment:
                case eToken_newline:
                  theIndex++;
                  break;
                default:
                  if(0==theString.CompareWithConversion("->",PR_TRUE,2)) {
                    theIndex=anIndex-1; //now start looking backwards...
                    theState=eFindIf;
                  }
                  else theState=eFailure;
              }//switch
              break;

          case eFindIf:
            switch(theType) {
              case eToken_whitespace:
              case eToken_comment:
              case eToken_newline:
                theIndex--;
                break;
              case eToken_semicolon:
                return true;
                break;

              case eToken_operator:
                switch(theChar){
                  case '{':
                  case '+':
                  case '-':
                  case '=':
                  case '*':
                  case '%':
                    return true;
                  default:
                    break;
                }
                theIndex-=1;
                break;
              case eToken_identifier:
                if(0==theString.CompareWithConversion("if",true,2)){
                  return false;
                }
                else if(0==theString.CompareWithConversion(aName)){
                  return false;
                }
                else if(0==theString.CompareWithConversion("while",true,5)){
                  return false;
                }
                return true;
              default:
                theIndex--;
            } //switch
            break;
          case eFailure:
            break;
        }//switch
      }//if
      else theState=eFailure;
    }//while

    return result;
  }

  void skipSemi(CCPPTokenizer& aTokenizer) {
    CToken*   theToken=0;
    while((theToken=aTokenizer.getTokenAt(++mIndex))){
      int     theType=theToken->getTokenType();
      nsCString& theString=theToken->getStringValue();
      PRUnichar   theChar=theString.GetBuffer()[0];

      switch(theType){
        case eToken_semicolon:
          return;
        case eToken_operator:
          if(('}'==theChar) || (';'==theChar))
            return;
          break;
        case eToken_newline:
        case eToken_comment:
        case eToken_whitespace:
          mLineNumber+=theString.CountChar('\n');
        default:
          break;
      } //switch
    }//while
  }

  void OutputMozillaFilename(ostream& anOutput) {
    
    PRInt32 pos=mFilename.RFind("mozilla");
	mFilename.ReplaceChar('\\','/');
    const char* buf=mFilename.GetBuffer();
    if(-1<pos) {
      buf+=pos+8;
  
      if(gEmitHTML) {
        anOutput << "<br><a href=\"" ;
      
        anOutput << "http://lxr.mozilla.org/mozilla/source/";
        anOutput << buf << "#" << mLineNumber;

        cout << "\">";  
      }

      cout << buf << ":" << mLineNumber;

      if(gEmitHTML) {
        cout << "</a>";
      }

    }
    else {
      anOutput << buf;
    }
  }


  /***************************************************************************
    This method iterates the tokens (in tokenizer) looking for statements.
    The statements are then decoded to find pointer variables. If they get a 
    new call, then they're recorded in mNames. If they're in an IF(), then
    they're recording in mSafeNames. We also look for simple dereferences.
    When we find one, we look it up in mSafeNames. If it's not there, we have
    a potential deref error.
   ***************************************************************************/
  void scan(CCPPTokenizer& aTokenizer,ostream& anOutput){
    int     theMax=aTokenizer.getCount();
    int     theIDCount=mNames.GetSize();
    int     theSafeIDCount=mSafeNames.GetSize();

    while(mIndex<theMax){
      CToken*   theToken=aTokenizer.getTokenAt(mIndex);
      int     theType=theToken->getTokenType();
      nsCString& theString=theToken->getStringValue();

      switch(theType){
        case eToken_operator:
          {
            PRUnichar   theChar=theString.GetBuffer()[0];
            switch(theChar){
              case '{':
                mIndex++;
                scan(aTokenizer,anOutput);
                break;
              case '}':
                //before returning, let's remove the new identifiers...
                while(mNames.GetSize()>theIDCount){
                  mNames.Pop();
                }
                while(mSafeNames.GetSize()>theSafeIDCount){
                  mSafeNames.Pop();
                }
                return;
              default:
                break;
            }
          }
          break;

        case eToken_comment:
        case eToken_newline:
        case eToken_whitespace:
          mLineNumber+=theString.CountChar('\n');
          break;

            //If it's an ID, then we're looking for 3 patterns:
            //  1. ID* ID...; 
            //  2. *ID=ID;
            //  3. ID->XXX;
        case eToken_identifier:
          {
            PRUnichar theOpChar=0;
            if((theToken=findPtrDecl(aTokenizer,mIndex,theOpChar,true))) {
              //we found ID* ID; (or) ID* ID=...;
              mNames.Push(theToken);
              mIndex+=2;
              skipSemi(aTokenizer);
            }

            else if(0==theString.CompareWithConversion("if",true,2)) {
              CToken* theSafeID=findSafeIDInIfStatement(aTokenizer);
              if(theSafeID) {
                mSafeNames.Push(theSafeID);
                theSafeIDCount++;
              }
            }
            else if(0==theString.CompareWithConversion("NS_ENSURE",true,9)) {
				CToken* theSafeID=findSafeIDInEnsureTrue(aTokenizer);
				if(theSafeID) {
					mSafeNames.Push(theSafeID);
					theSafeIDCount++;
					}
				}
			else if(!gSloppy && (theString.Length()==3) && (0==theString.CompareWithConversion("new",true,3))) {
				CToken* theToken=findNewVariable(aTokenizer,mIndex);
				if(theToken) {
					mNames.Push(theToken);
					theIDCount++;
					CToken* lastSafe=(CToken*)mSafeNames.PeekFront();
					if(lastSafe){
						if(theToken==lastSafe){
							mSafeNames.Pop();
							theSafeIDCount--;
						}
					}
				}
			}
			else {
              CToken* theMatch=matchID(mNames,theString);
              if(theMatch){
                if((hasUnconditionedDeref(theString,aTokenizer,mIndex)) || 
                   (hasSimpleDeref(aTokenizer,mIndex))){
                  CToken* theSafeID=matchID(mSafeNames,theString);
                  if(theSafeID) {
                    nsCString& s1=theSafeID->getStringValue();
                    if(s1.Equals(theString))
                      break;
                  }
                  //dump the name out in LXR format

                  OutputMozillaFilename(anOutput);
                  nsCAutoString theCStr(theString);


                  anOutput << "   Deref-error: \"" << (const char*)theCStr << "\"" << endl;
                  mErrCount++;
                }
              }
			}
          }
          break;
        default:
          break;
      }
      mIndex++;
    }
  }
  
  nsDeque     mNames;
  nsDeque     mSafeNames;
  int         mIndex;
  int         mLineNumber;
  int         mErrCount;
  nsCString   mFilename;
};

void ScanFile(nsCString& aFilename,ostream& anOutput,int& aLineCount,int& anErrCount) {
  nsCAutoString theCStr(aFilename);
 
  fstream input((const char*)theCStr,ios::in);
  CScanner theScanner(input);
  CCPPTokenizer theTokenizer;
  theTokenizer.tokenize(theScanner);

  CPattern thePattern(aFilename);
  thePattern.scan(theTokenizer,anOutput);
  aLineCount+=thePattern.mLineNumber;
  anErrCount+=thePattern.mErrCount;
}


void IteratePath(const char *aPath, ostream& anOutput,
  int& aFilecount,int& aLinecount,int& anErrcount) {

  PRDir *aDir;

  if ((aDir = PR_OpenDir(aPath))) {
    PRDirEntry *anEntry;
    while ((anEntry = PR_ReadDir(aDir, PR_SKIP_BOTH))) {
      char buf[256];
      CBufDescriptor bud(buf, PR_TRUE, sizeof buf, 0);
      nsCAutoString path(bud);
      PRFileInfo eInfo;

      path = aPath; path += '/'; path += anEntry->name;
      if (PR_SUCCESS == PR_GetFileInfo(path.GetBuffer(), &eInfo)) {
        if (eInfo.type == PR_FILE_DIRECTORY) {
          if (0 != PL_strcasecmp(anEntry->name, "CVS"))
            IteratePath(path.GetBuffer(), anOutput, aFilecount, aLinecount, anErrcount);
        }
        else if (eInfo.type == PR_FILE_FILE) {
          const char *ext = PL_strcaserstr(anEntry->name, ".cpp");
          if (ext && (ext[4] == 0)) {
            aFilecount++;
            ScanFile(path,anOutput,aLinecount,anErrcount);
          }
        }
      }
    }
  }
}
 

void ConsumeArguments(int argc, char* argv[]) {

  for(int index=1;index<argc;index++){
    switch(argv[index][0]) {
      case '-':
      case '/':
        switch(argv[index][1]){
          case 'd':
          case 'D': // Starting directory
            gRootDir = argv[index+1];
            index++;
            break;

          case 'h':
          case 'H': // Output HTML with lxr.mozilla.org links
            gEmitHTML=true;
            break;
		  case 's':
		  case 'S':
			  gSloppy=true; // Dont look for member variables
			  break;

          default:
            cout << "Usage: dreftool [-h] [-d sourcepath]" << endl;
            cout << "-d path to root of source tree" << endl;
			cout << "-s sloppy mode no warnings if member variables are used unsafe" <<endl;
            cout << "-h emit as HTML" << endl;
            exit(1);
        }
        break;
      default:
        break;
    }
  }
}

int main(int argc, char* argv[]){

  gRootDir = ".";
  ConsumeArguments(argc,argv);

  int fileCount=0;
  int lineCount=0;
  int errCount=0;

#if 0
  nsCAutoString temp("test3.cpp");
  ScanFile(temp,cout,lineCount,errCount);
  fileCount++;
#endif

#if 1

  if(gEmitHTML) {
    cout << "<html><body>" << endl;
  }

  IteratePath(gRootDir,cout,fileCount,lineCount,errCount);
#endif

  if(gEmitHTML) {
    cout << "<PRE>" <<endl;
  }

  cout << endl << "Summary: " << endl;
  cout << "===============================" << endl;
  cout<< "Files:  " << fileCount << endl;
  cout<< "Lines:  " << lineCount<< endl;
  cout<< "Errors: " << errCount<< endl;

  if(gEmitHTML) {
    cout << "</PRE>" <<endl;
    cout << "</body></html>" << endl;
  }


  return 0;
}
